__version__ = '0.4.1'
__author__ = 'elegans.io Ltd'
__license__ = 'GPL v2'
__maintainer__ = 'elegans-io'
__email__ = 'info@elegans.io'

from csrec.factory_dal import Dal
from csrec.recommender import Recommender
